# -*- coding: utf-8 -*-
"""

"""
import numpy as np
from scipy.optimize import minimize_scalar
from ..special import u0, eps0, c0, Nm_to_dBcm, eta0


class Waveguide:
    """
    Create an SIW object with a given geometry
    """

    def __init__(self, _metal, _diel, _height):
        self.metal = _metal
        self.diel = _diel
        self.width = 0.0
        self.first_cut_off = 0.0
        self.height = _height
        self.eta = np.sqrt(u0 / (self.diel.epsilon * eps0))

    @property
    def width(self) -> float:
        if self.f_c <= 0.0:
            return self._w
        self._w = c0 / (self.f_c * 2 * np.sqrt(self.diel.epsilon))
        return self._w

    @width.setter
    def width(self, _width):
        """
        set the width of the wave-guide and update the cut-off frequency
        """
        self._w = _width

    @property
    def first_cut_off(self):
        self.f_c = self.f_cut_off()
        return self.f_c

    @first_cut_off.setter
    def first_cut_off(self, _fc):
        """
        set the cut-off frequency of the wave-guide and update the width
        """
        self.f_c = _fc

    def f_cut_off(self, _m=1, _n=0):
        """
        return the value of the cut-off frequency of the TEM mode _m, _n
        """
        eps = self.diel.epsilon
        return (
            c0
            * np.sqrt((_m * np.pi / self.width) ** 2 + (_n * np.pi / self.height) ** 2)
            / (2 * np.pi * np.sqrt(eps))
        )

    def calc_k(self, _freq):
        """
        convert the freq in pulsation in the given substrate
        """
        return np.sqrt(self.diel.epsilon) * 2 * np.pi * _freq / c0

    def calc_lambda(self, _f):
        """
        return the wavelength inside the waveguide at the given frequency _f
        """
        return c0 / (_f * np.sqrt(1 - (self.f_c / _f) ** 2))

    def calc_beta(self, _freq):
        """
        return the value of the velocity (beta)
        """
        return np.sqrt(self.calc_k(_freq) ** 2 - (np.pi / self.width) ** 2)

    def calc_a_d(self, _freq):
        """
        return the value of the dielectric loss in dB/m at the frequency freq (array-like)
        """
        k = self.calc_k(_freq)
        tan_d = self.diel.tan_d
        beta = self.calc_beta(_freq)
        return Nm_to_dBcm * k ** 2 * tan_d / (2 * beta)

    def calc_a_c(self, _freq):
        """
        return the value of the conductor loss in dB/m at the frequency freq (array-like)
        """
        r_s = np.sqrt(2 * np.pi * _freq * u0 / (2 * self.metal.rho))
        eta = self.eta
        k = self.calc_k(_freq)
        beta = self.calc_beta(_freq)
        height = self.height
        width = self.width
        return (
            Nm_to_dBcm
            * r_s
            * (2 * height * np.pi ** 2 + width ** 3 * k ** 2)
            / ((width ** 3) * height * beta * k * eta)
        )

    def calc_ksr(self, _freq):
        """
        return the coefficient of the added conductor loss introduce by surface roughness
        """
        rho = self.metal.rho
        skin_d = 1 / np.sqrt(rho * np.pi * _freq * u0)
        rough = self.diel.roughness
        if rough <= 0:
            raise ValueError(
                "Roughness must be above zero. \
Value can be set through /self.diel.roughness/"
            )
        return 2 * np.arctan(1.4 * (rough / skin_d) ** 2) / np.pi

    def calc_pphc(self, _freq, _e_0):
        """
        return the peak power handling capability in watt
        at the _freq frequency (in GHz) and for a maximum electric field _e_0 (in V/m)
        """
        width = self.width
        if width <= 0.0:
            raise ValueError(
                "Width must be above zero. \
Value can be set using set_width() or set_f_c()"
            )
        height = self.height
        f_c = self.first_cut_off
        eps = self.diel.epsilon
        return (
            0.25
            * np.sqrt(eps)
            * np.sqrt(1 - (f_c / _freq) ** 2)
            * width
            * height
            * _e_0 ** 2
            / eta0
        )

    def calc_aphc(self, _freq, _t_max, t_amb=295) -> float:
        # using doi.org/10.1109/TADVP.2008.927814
        a = self.width
        alpha = self.calc_a_c(_freq) / Nm_to_dBcm + self.calc_a_d(_freq) / Nm_to_dBcm
        # see page 346 of Heat Transfert, 10th edition
        h_uc = 1.32 * ((_t_max - t_amb) / a) ** 0.25
        h_dc = 0.59 * ((_t_max - t_amb) / a) ** 0.25
        # see page 461 of Heat Transfert, 10th edition (with A2 >> A1)
        sigma = 5.669 * 1e-8  # W/m²K^4 Stefan-Boltzmann constant
        epsilon = 0.65  # emissivity of the copper
        h_r = sigma * epsilon * (_t_max ** 2 + t_amb ** 2) * (_t_max + t_amb)
        return (h_uc + h_dc + 2 * h_r) * a * (_t_max - t_amb) / alpha

    def print_info(self):
        """
        output the size and the upper mode cut-off frequency
        """
        fc_01 = self.f_cut_off(0, 1)
        print(f"Width: {self.width * 1e3:.2f} mm\tfc01: {fc_01 * 1e-9:.2f} GHz")

    def get_sparam(self, _freq, _length):
        """
        return the 4 scattering parameters of a wave-guide section
        of the given length for the given frequency
        """
        s11 = 0
        alpha = self.calc_a_c(_freq) + (1 + self.calc_ksr(_freq)) * self.calc_a_d(_freq)
        s21 = (1 - s11) * np.exp(-(alpha + 1j * self.calc_beta(_freq)) * _length)
        return s21


class AF_SIW(Waveguide):
    """
    Create an AF-SIW object with a given geometry
    """

    def __init__(self, _metal, _diel, _height, _slab):
        self.slab = _slab
        Waveguide.__init__(self, _metal, _diel, _height)
        self.width = 0.0
        self.first_cut_off = 0.0

    @property
    def width(self) -> float:
        if self.f_c > 0:
            slb = self.slab
            sqr_eps = np.sqrt(self.diel.epsilon)
            tan = np.tan(2 * slb * np.pi * self.f_c / c0) * sqr_eps
            self._w = 2 * slb + np.arctan(1 / tan) * c0 / (sqr_eps * np.pi * self.f_c)
        return self._w

    @width.setter
    def width(self, _width):
        """
        set the width of the wave-guide and update the cut-off frequency
        """
        self._w = _width

    @property
    def slab(self) -> float:
        return self._slab

    @slab.setter
    def slab(self, _s: float):
        if _s <= 0:
            raise ValueError("Slab must be above zero. Please use Waveguide class")
        self._slab = _s

    @property
    def first_cut_off(self) -> float:
        if self.width > 0:
            self.f_c = self.f_cut_off(1, 0)
        return self.f_c

    @first_cut_off.setter
    def first_cut_off(self, _fc):
        """
        set the cut-off frequency of the wave-guide and update the width
        """
        self.f_c = _fc

    def f_cut_off(self, _m=1, _n=0) -> float:
        """
        return the value of the cut-off frequency of the TEM mode _m, _n
        """
        if _n > 0:
            raise ValueError("Value of _n greater than 0 are not supported")
        if _m % 2 == 1:
            res = minimize_scalar(self.__odd_fc)
        else:
            f_c = self.f_c
            res = minimize_scalar(
                self.__even_fc, bounds=(1.5 * f_c, 5 * f_c), method="bounded"
            )
        return res.x

    def __odd_fc(self, _fc):
        if _fc <= 0:  # frequency must be strictly positive
            return 1e12
        slb = self._slab
        wth = self._w
        sqr_eps = np.sqrt(self.diel.epsilon)
        return np.abs(
            sqr_eps * np.tan(2 * np.pi * _fc * slb / c0)
            - 1 / np.tan(sqr_eps * np.pi * _fc * (wth - 2 * slb) / c0)
        )

    def __even_fc(self, _fc):
        slb = self._slab
        wth = self._w
        sqr_eps = np.sqrt(self.diel.epsilon)
        return np.abs(
            sqr_eps * np.tan(2 * np.pi * _fc * slb / c0)
            + np.tan(sqr_eps * np.pi * _fc * (wth - 2 * slb) / c0)
        )

    def calc_a_d(self, _freq):
        return 0

    def calc_ksr(self, _freq):
        rho = self.metal.rho
        skin_d = 1 / np.sqrt(rho * np.pi * _freq * u0)
        rough = self.metal.roughness
        return 2 * np.arctan(1.4 * (rough / skin_d) ** 2) / np.pi

    def print_info(self):
        """
        output the size and the upper mode cut-off frequency
        """
        fc_20 = self.f_cut_off(2, 0)
        print(f"Width: {self.width * 1e3:.2f} mm\tfc20: {fc_20 * 1e-9:.2f} GHz")



"""
Implementation of an aggressive space mapping algorithm for RF-design
"""
import functools
from scipy.optimize import minimize


def space_map(coarse_model, dim0, fine_model, par0, goal, maxiter=5):
    """
    Optimization function for space mapping algorithm.

    Parameters
    ----------
    coarse_model : fun
        function evaluating the coarse model goals
        for a set of dimensions (dim) and parameters (par)
    dim0 : dict
        initial dimensions of the component
    fine_model : fun
        function evaluating the fine model goals
        for a set of dimensions (dim)
    par0 : dict
        initial parameters of the component coarse model
    goal : dict
        set of goal targeted by the algorithm
    maxiter : int, optional
        maximal number of iteration. The default is 5.

    Returns
    -------
    dim : dict
        final dimension of the component.
    par : dict
        final parameters of the component model.
    fine_mod : dict
        achieved goal.

    """
    dim = dim0
    par = par0
    achieved_goal = goal
    for i in range(maxiter):
        # evaluate exact value using fine model
        achieved_goal = fine_model(dim)
        # alter coarse model parameters to better match fine model results
        res = minimize(
            __refresh_coarse,
            __totuple(par),
            method="L-BFGS-B",
            args=(par0.keys(), dim, coarse_model, achieved_goal),
        )
        par = __todict(par0.keys(), res.x)
        # find best solution according to coarse model
        res = minimize(
            __cost_coarse,
            __totuple(dim),
            method="Powell",
            args=(dim0.keys(), par, coarse_model, goal),
        )
        dim = __todict(dim0.keys(), res.x)
    return dim, par, achieved_goal


def __cost_coarse(dim_values, dim_keys, par, coarse_model, goal):
    _dim = __todict(dim_keys, dim_values)
    coarse_mod = coarse_model(_dim, par)
    return cost_calc(__totuple(coarse_mod), __totuple(goal))


def __refresh_coarse(par_values, par_keys, dim, coarse_model, fine_mod):
    _par = __todict(par_keys, par_values)
    coarse_mod = coarse_model(dim, _par)
    return cost_calc(__totuple(coarse_mod), __totuple(fine_mod))


@functools.lru_cache()
def cost_calc(perf_list, goal_list, weight_list=None):
    """
    return the normalize standard deviation between the perf_list and the goal_list

    Parameters
    ----------
    perf_list : tuple of float
        list of the performances achieved.
    goal_list : tuple of float
        list of the goal to be achieved. If one value is given, the goal is a point. If two, the goal is an interval.
    weight_list : tuple of float, optional
        weightning of the goals. If set to None, all the weight are set to one.

    Returns
    -------
    cost : float
        cost value

    """
    if weight_list is None:
        weight_list = [1] * len(goal_list)
    cost = 0
    for itern, goal in enumerate(goal_list):
        perf = perf_list[itern]
        goal = goal_list[itern]
        if isinstance(goal, tuple):
            err_min = max((perf - goal[0]) / (goal[0] if goal[0] != 0 else 1), 0)
            err_max = max((goal[1] - perf) / (goal[1] if goal[1] != 0 else 1), 0)
            err = err_min + err_max
        else:
            err = (perf - goal) / (goal if goal != 0 else 1)
        cost += weight_list[itern] * err ** 2
    return cost


def __todict(keys, values):
    return dict(zip(keys, values))


def __totuple(dicts):
    return tuple(dicts.values())



# -*- coding: utf-8 -*-
"""

"""
import numpy as np
import yaml
from scipy.optimize import minimize, OptimizeResult
from ..components.transformer import Transformer
from ..special import std_dev, quality_f


class Balun:
    """
    Create a balun object
    """

    def __init__(self, _fc=1e9, _z_source=50, _z_load=50, modelmapfile=None):
        self.f_c = _fc
        self.z_src = _z_source
        self.z_ld = _z_load
        self.is_symmetrical = False
        if modelmapfile is None:
            modelmapfile = "tests/default.map"
        with open(modelmapfile, "r") as file:
            self.modelmap = yaml.full_load(file)
        width_lim = (
            float(self.modelmap["width"]["min"]),
            float(self.modelmap["width"]["max"]),
        )
        turn_lim = (
            float(self.modelmap["turn"]["min"]),
            float(self.modelmap["turn"]["max"]),
        )
        gap_lim = float(self.modelmap["gap"])
        self.bounds = np.array(
            [
                width_lim,  # width
                turn_lim,  # turn number
                (width_lim[1], 1e6),  # inner diameter
                (gap_lim, 1e6),  # gap
                width_lim,
                turn_lim,
                (width_lim[1], 1e6),
                (gap_lim, 1e6),
            ]
        )
        geo = {
            "di": 20,
            "n_turn": 1,
            "width": 2e-6,
            "gap": 2e-6,
        }
        self.transfo = Transformer(geo, geo, _fc, modelmapfile)

    def __cost_geo_vs_targ(self, geo, _l_targ, _is_primary=True):
        """
        return the cost (standard deviation)
        between the proposed solution and the targeted specifications
        """
        if _is_primary:
            self.transfo.set_primary(
                {
                    "di": geo[2],
                    "n_turn": np.round(geo[1]),
                    "width": geo[0],
                    "gap": geo[3],
                }
            )
            l_sol = self.transfo.model["lp"]
            r_sol = self.transfo.model["rp"]
        else:
            self.transfo.set_secondary(
                {
                    "di": geo[2],
                    "n_turn": np.round(geo[1]),
                    "width": geo[0],
                    "gap": geo[3],
                }
            )
            l_sol = self.transfo.model["lp"]
            r_sol = self.transfo.model["rp"]
        return std_dev(l_sol, _l_targ) + np.sum(r_sol) / 100

    def design(self, _maxiter=1000):
        """
        design an impedance transformer
        with the targeted specifications (f_targ, zl_targ, zs_targ)
        return an optimization results (res)
        """
        k = self.transfo.model["k"]
        alpha = (1 - k ** 2) / k
        q_s = -quality_f(self.z_src)
        q_l = -quality_f(self.z_ld)
        # assuming perfect inductor for first calculation
        r_l1 = 0
        r_l2 = 0
        for i in range(2):
            q_s_prime = q_s * np.real(self.z_src) / (np.real(self.z_src) + r_l1)
            q_l_prime = q_l * np.real(self.z_ld) / (np.real(self.z_ld) + r_l2)
            b_coeff = 2 * alpha * q_s_prime + q_s_prime + q_l_prime
            discr = b_coeff ** 2 - 4 * alpha * (alpha + 1) * (1 + q_s_prime ** 2)
            if discr < 0:
                raise ValueError(
                    "Negative value in square root,\
 try to increase the coupling factor or the load quality factor\
or try to lower the source quality factor"
                )
            z_sol = np.array(
                (
                    (b_coeff + np.sqrt(discr)) / (2 * (alpha + 1)),
                    (b_coeff - np.sqrt(discr)) / (2 * (alpha + 1)),
                )
            )
            qxl1 = z_sol / (1 - k ** 2)
            qxl2 = (
                z_sol * (1 + q_l_prime ** 2) / (alpha * (1 + (q_s_prime - z_sol) ** 2))
            )
            l_sol1 = qxl1 * np.real(self.z_src) / (2 * np.pi * self.f_c)
            l_sol2 = qxl2 * np.real(self.z_ld) / (2 * np.pi * self.f_c)
            if l_sol1[0] * l_sol2[0] > l_sol1[1] * l_sol2[1]:
                # selecting the solution giving the smallest inductors
                l_1 = l_sol1[1]
                l_2 = l_sol2[1]
            else:
                l_1 = l_sol1[0]
                l_2 = l_sol2[0]
            # find the inductor geometry that give the desired inductances
            res1 = minimize(
                fun=self.__cost_geo_vs_targ,
                x0=[x for x in self.transfo.prim.values()],
                bounds=self.bounds[0:4],
                args=l_1,
            )
            if self.is_symmetrical:
                self.__cost_geo_vs_targ(res1.x, l_2, _is_primary=False)
                res2 = res1
            else:
                res2 = minimize(
                    fun=self.__cost_geo_vs_targ,
                    x0=[x for x in self.transfo.prim.values()],
                    bounds=self.bounds[4:],
                    args=(l_2, False),
                )
            r_l1 = self.transfo.model["rs"]
            r_l2 = self.transfo.model["rp"]
        res = OptimizeResult()
        res.x = np.concatenate((res1.x, res2.x))
        res.fun = (res1.fun + res2.fun) / 2
        res.message = res1.message
        return res

    def __enforce_symmetrical(self, _q_val, _of_load=True):
        """
        return the 'distance' to a symmetrical balun (ie. primary = secondary)
        if _of_load, altering the load impedance
        else altering the source impedance
        """
        k = self.transfo.model["k"]
        alpha = (1 - k ** 2) / k
        if _of_load:
            q_s = -quality_f(self.z_src)
            q_l = _q_val
        else:
            q_s = _q_val
            q_l = -quality_f(self.z_ld)
        b_coeff = 2 * alpha * q_s + q_s + q_l
        discr = b_coeff ** 2 - 4 * alpha * (alpha + 1) * (1 + q_s ** 2)
        if discr < 0:
            return np.inf
        z_sol = np.array(
            (
                (b_coeff + np.sqrt(discr)) / (2 * (alpha + 1)),
                (b_coeff - np.sqrt(discr)) / (2 * (alpha + 1)),
            )
        )
        qxl1 = z_sol / (1 - k ** 2)
        qxl2 = z_sol * (1 + q_l ** 2) / (alpha * (1 + (q_s - z_sol) ** 2))
        qxl_ratio = np.real(self.z_ld) / np.real(self.z_src)
        return np.abs(np.min(qxl1 / qxl2) - qxl_ratio)

    def enforce_symmetrical(self, side="load", _verbose=True):
        """
        alter the value of the load impedance (if side="load") or the source impedance
        in order to realize a symmetrical balun (ie. primary = secondary)
        """
        if side == "load":
            old_z = self.z_ld
            _through_load = True
        else:
            old_z = self.z_src
            _through_load = False
        res = minimize(
            self.__enforce_symmetrical,
            -quality_f(old_z),
            args=_through_load,
            method="Nelder-Mead",
        )
        new_z = np.real(old_z) * (1 - 1j * res.x)
        if _verbose:
            print(f"old z_ld: ${complex(old_z):5.2f}")
            print(f"new z_ld: ${complex(new_z):5.2f}")
        if _through_load:
            self.z_ld = new_z
        else:
            self.z_src = new_z
        self.is_symmetrical = True

    def print(self, res):
        """
        print a summary of the solution (res)
        with a comparison to the boundaries
        """
        sol = res.x * 1e6
        bds = np.array(self.bounds) * 1e6
        print(f"Solution funds with remaining error of: {float(res.fun):.2e}")
        print("Termination message of algorithm: " + str(res.message))
        print("\t\t\tW (µm)\tn\tdi (µm)\tG (µm)")
        print(
            f"lower bound :\t{(bds[0])[0]:.2g}\t{(self.bounds[1])[0]:.2g}\t\
{(bds[2])[0]:.3g}\t{(bds[3])[0]:.2g}"
        )
        print(
            f"primary dim.:\t{sol[0]:.2g}\t{res.x[1]:.0g}\t{sol[2]:.3g}\t{sol[3]:.2g}"
        )
        print(
            f"secondary dim.:\t{sol[4]:.2g}\t{res.x[5]:.0g}\t{sol[6]:.3g}\t{sol[7]:.2g}"
        )
        print(
            f"upper bound :\t{(bds[0])[1]:.2g}\t{(self.bounds[1])[1]:.2g}\t\
{(bds[2])[1]:.3g}\t{(bds[3])[1]:.2g}"
        )



# -*- coding: utf-8 -*-
"""

"""
import yaml
import numpy as np
from scipy.optimize import minimize, minimize_scalar, OptimizeResult
from ..components.transformer import Transformer
from ..special import ihsr


class Coupler:
    """
    Create a coupler object
    """

    def __init__(self, _fc=1e9, _zc=50, modelmapfile=None, bounds=None):
        self.f_c = _fc
        self.z_c = _zc
        if modelmapfile is None:
            modelmapfile = "tests/default.map"
        with open(modelmapfile, "r") as file:
            self.modelmap = yaml.full_load(file)
        width_lim = (
            float(self.modelmap["width"]["min"]),
            float(self.modelmap["width"]["max"]),
        )
        turn_lim = (
            float(self.modelmap["turn"]["min"]),
            float(self.modelmap["turn"]["max"]),
        )
        gap_lim = float(self.modelmap["gap"])
        if bounds is None:
            self.bounds = np.array(
                [
                    width_lim,  # width
                    turn_lim,  # turn number
                    (width_lim[1], 20 * width_lim[1]),  # inner diameter
                    (gap_lim, 1.01 * gap_lim),  # gap
                ]
            )
        else:
            self.bounds = bounds
        geo = {
            "di": 20,
            "n_turn": 2,
            "width": self.bounds[0, 0],
            "gap": self.bounds[3, 0],
        }
        self.transfo = Transformer(geo, geo, _fc, modelmapfile)

    def cost(self, sol):
        """
        return the cost (standard deviation)
        between the proposed solution and the targeted specifications
        """
        geo = {
            "di": sol[2],
            "n_turn": np.round(sol[1]),
            "width": sol[0],
            "gap": sol[3],
        }
        self.transfo.set_primary(geo)
        self.transfo.set_secondary(geo)
        s_p = self.transfo.circuit.s_external[0]
        return np.abs(s_p[0, 0]) + np.max([26.7 - ihsr(s_p[1, 0], s_p[0, 1]), 0])

    def __cost_est_inductance(self, _di):
        self.transfo.prim["di"] = _di
        return np.abs(self.transfo.l_geo() - self.z_c / (2 * np.pi * self.f_c))

    def __cost_est_capacitance(self, _width):
        self.transfo.prim["width"] = _width
        return np.abs(
            self.transfo.cc_geo()
            + self.transfo.cc_geo(False)
            - 1 / (self.z_c * 2 * np.pi * self.f_c)
        )

    def design(self, _maxiter=500):
        """
        design an hybrid coupleur with the targeted specifications (f_targ, z_targ)
        return an optimization results (res)
        """
        # finding the inner diameter that give the correct inductance
        minimize_scalar(self.__cost_est_inductance, bounds=self.bounds[2])
        # finding the path width that give the correct capacitor
        res_int = minimize_scalar(self.__cost_est_capacitance, bounds=self.bounds[0])
        geo = self.transfo.prim
        x_0 = np.array([geo["width"], geo["n_turn"], geo["di"], geo["gap"]])
        if _maxiter == 0:  # just get the first guess
            res = OptimizeResult()
            res.x = x_0
            res.fun = res_int.fun
            res.message = "First Guess"
            return res
        res = minimize(self.cost, x0=x_0)
        return res

    def print(self, res):
        """
        print a summary of the solution (res)
        with a comparison to the boundaries
        """
        sol = res.x * 1e6
        bds = np.array(self.bounds) * 1e6
        print(f"Solution funds with remaining error of: {float(res.fun):.2e}")
        print("Termination message of algorithm: " + str(res.message))
        print("\t\tW (µm)\tn\tdi (µm)\tG (µm)")
        print(
            f"lower bound :\t{(bds[0])[0]:.2g}\t{(self.bounds[1])[0]:.2g}\t\
{(bds[2])[0]:.3g}\t{(bds[3])[0]:.2g}"
        )
        print(
            f"best point  :\t{sol[0]:.2g}\t{res.x[1]:.0g}\t{sol[2]:.3g}\t{sol[3]:.2g}"
        )
        print(
            f"upper bound :\t{(bds[0])[1]:.2g}\t{(self.bounds[1])[1]:.2g}\t\
{(bds[2])[1]:.3g}\t{(bds[3])[1]:.2g}"
        )
